#include <stdio.h>
#include "Gamebuino-Meta.h"

#include "room.h"

uint8_t MIN_SIZE = 5;

Room::Room(uint8_t top, uint8_t left, uint8_t width, uint8_t height) :
    top(top),
    left(left),
    width(width),
    height(height)
    {}

Room::~Room() {
    if (type == NODE) {
        delete child[0];
        child[0] = 0;
        delete child[1];
        child[1] = 0;
    }
}

bool Room::split() {

    if (type == NODE) {
        uint8_t leftSplit = child[0]->split();
        uint8_t rightSplit = child[1]->split();
        return (leftSplit || rightSplit);
    }

    bool horizontal = (rand() % 2);
    /* maximum height/width we can split off */
    uint8_t max = (horizontal ? height : width ) - MIN_SIZE; 

    if (max <= MIN_SIZE) {
        return false;
    }
    uint8_t split = rand() % max;
    if (split < MIN_SIZE) split = MIN_SIZE;

    if (horizontal) {
        child[0]  = new Room(top, left, width, split); 
        child[1] = new Room(top+split, left, width, height-split );
    } else {
        child[0]  = new Room(top, left, split, height);
        child[1] = new Room(top, left+split, width-split, height);
    }
    type = NODE;
    return true;
}

bool Room::generate(uint8_t maxDepth) {
    for (uint8_t i = 0; i < maxDepth; i++) {
        if (!split())
            break;
    }

    if (type == LEAF) {
        return false;
    }

    addDoors();


    printf("There are %d rooms\n", count());
    return true;
}

bool Room::addDoor(Room* room) {
    uint8_t idx_1, idx_2;

    if ((doors[2] != 0) || room->doors[2] != 0) {
        return false;
    }

    for (idx_1 = 0; idx_1 < 3; idx_1++) {
        Room* target = doors[idx_1];
        if (target == 0) {
            break;
        }
        if (target == room) {
            return false;
        }
    }

    for (idx_2 = 0; idx_2 < 3; idx_2++) {
        if (room->doors[idx_2] == 0) {
            break;
        }
    }

    doors[idx_1] = room;
    room->doors[idx_2] = this;
    return true;
}

void Room::addDoors() {

    if (type == LEAF) {
        return;
    }

    if (child[0]->type == LEAF && child[1]->type == LEAF) {
        child[0]->addDoor(child[1]);
    } else if (child[0]->type == LEAF) {
        Room* min = child[1];
        while (min->type == NODE) {
            min = min->child[0];
        }
        child[0]->addDoor(min);
    } else if (child[1]->type == LEAF) {
        Room* max = child[0];
        while (max->type == NODE) {
            max = max->child[1];
        }
        child[1]->addDoor(max);
    } else {
        Room* tmpRoom;
        for (uint8_t iter = 0; iter < 2; iter++) {
            Room* target = child[0];
            Room* targetChild = child[1];
            while (targetChild->type == NODE) {

                bool connected;
                if (iter) {
                    /*
                     * Check if two rooms are connected by top left
                     */
                    connected = 
                        (target->top == targetChild->child[0]->top && 
                         ((target->left+target->width == targetChild->child[0]->left) || 
                          (target->left-targetChild->child[0]->width == targetChild->child[0]->left)))
                    ||  (target->left == targetChild->child[0]->left && 
                            ((target->top+target->height == targetChild->child[0]->top) || 
                             (target->top-targetChild->child[0]->height == targetChild->child[0]->top)));
                
                } else {
                    /*
                     * Check if two rooms are connected by bottom right.
                     */
                    connected =
                        (target->top + target->height == targetChild->child[0]->top + targetChild->child[0]->height && 
                         ((target->left+target->width == targetChild->child[0]->left) || 
                          (target->left-targetChild->child[0]->width == targetChild->child[0]->left)))
                    ||  (target->left + target->width == targetChild->child[0]->left + targetChild->child[0]->width && 
                            ((target->top+target->height == targetChild->child[0]->top) || 
                             (target->top-targetChild->child[0]->height == targetChild->child[0]->top)));
                }


                if (connected) {
                    tmpRoom = targetChild->child[0];
                } else {
                    /*
                     * We do not test here that this child is connected : it is.
                     */
                    tmpRoom = targetChild->child[1];
                }

                if (tmpRoom->type == LEAF) {
                    /*
                     * If the room as no child, just swap it with our source
                     */
                    targetChild = target;
                    target = tmpRoom;
                } else {
                    /*
                     * Go deeper as long as we can
                     */
                    targetChild = tmpRoom;
                }
            }
            target->addDoor(targetChild);
        }

    }

    child[0]->addDoors();
    child[1]->addDoors();
}

void Room::repr(FILE * pFile) {
    fprintf(pFile, "digraph G {\nnode [shape=box]\nedge [arrowhead=\"box\" arrowtail=\"box\"]\n");
    _repr(pFile);
    fprintf(pFile, "\n}");
}


void Room::_repr(FILE * pFile) {

    static uint8_t id = 0;
    switch (type) {
        case NODE:
            child[0]->_repr(pFile);
            child[1]->_repr(pFile);
            break;
        case LEAF:
            uint8_t i = 0;
            for (i = 0; i < 3; i++) {
                if (doors[i] == 0) {
                    break;
                }
                fprintf(pFile, "node_%p -> node_%p\n", this, doors[i]);
            }
            const char* color;
            if (i == 1) {
                color = "style=filled";
            } else {
                color = "";
            }

            fprintf(pFile, "node_%p [label= \"%d\" pos=\"%f,%f!\" width=\"%f\" height=\"%f\" %s]\n", \
                    this, id++, \
                    (left + ((float)width / 2))/5, \
                    (100 - top - ((float)height) / 2)/5, \
                    (float)width/5, \
                    (float)height/5, \
                    color);
    
    }
}

int Room::count() {
    if (type == LEAF) {
        return 1;
    } else {
        return child[0]->count() + child[1]->count();
    }
}
